/*
Copyright (c) 2014, Romain Dolbeau, unless otherwise noted.
No claims is made upon work by others.

For the work by Romain Dolbeau:
All rights reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

For the work by others:
See the indicated reference for the relevant license.
*/

#include <stdint.h>

#include "epi_crypto_common.h"

#include "e_epi_crypto.h"

#include "aes256_impl.h"

#ifdef __HOST__
#include <e-hal.h>
#include <string.h>
#include <openssl/evp.h>
#else
#include <e_lib.h>
#endif

#define f_FSb_32__1(x) 	((FSb[((x) >> 24) &0xFF] << 24) ^ \
                         (FSb[((x) >> 16) &0xFF] << 16))

#define f_FSb_32__2(x) 	((FSb[((x) >>  8) &0xFF] <<  8 ) ^ \
                         (FSb[((x)      ) &0xFF] & 0xFF))

#define FT \
    V(C6,63,63,A5), V(F8,7C,7C,84), V(EE,77,77,99), V(F6,7B,7B,8D), \
    V(FF,F2,F2,0D), V(D6,6B,6B,BD), V(DE,6F,6F,B1), V(91,C5,C5,54), \
    V(60,30,30,50), V(02,01,01,03), V(CE,67,67,A9), V(56,2B,2B,7D), \
    V(E7,FE,FE,19), V(B5,D7,D7,62), V(4D,AB,AB,E6), V(EC,76,76,9A), \
    V(8F,CA,CA,45), V(1F,82,82,9D), V(89,C9,C9,40), V(FA,7D,7D,87), \
    V(EF,FA,FA,15), V(B2,59,59,EB), V(8E,47,47,C9), V(FB,F0,F0,0B), \
    V(41,AD,AD,EC), V(B3,D4,D4,67), V(5F,A2,A2,FD), V(45,AF,AF,EA), \
    V(23,9C,9C,BF), V(53,A4,A4,F7), V(E4,72,72,96), V(9B,C0,C0,5B), \
    V(75,B7,B7,C2), V(E1,FD,FD,1C), V(3D,93,93,AE), V(4C,26,26,6A), \
    V(6C,36,36,5A), V(7E,3F,3F,41), V(F5,F7,F7,02), V(83,CC,CC,4F), \
    V(68,34,34,5C), V(51,A5,A5,F4), V(D1,E5,E5,34), V(F9,F1,F1,08), \
    V(E2,71,71,93), V(AB,D8,D8,73), V(62,31,31,53), V(2A,15,15,3F), \
    V(08,04,04,0C), V(95,C7,C7,52), V(46,23,23,65), V(9D,C3,C3,5E), \
    V(30,18,18,28), V(37,96,96,A1), V(0A,05,05,0F), V(2F,9A,9A,B5), \
    V(0E,07,07,09), V(24,12,12,36), V(1B,80,80,9B), V(DF,E2,E2,3D), \
    V(CD,EB,EB,26), V(4E,27,27,69), V(7F,B2,B2,CD), V(EA,75,75,9F), \
    V(12,09,09,1B), V(1D,83,83,9E), V(58,2C,2C,74), V(34,1A,1A,2E), \
    V(36,1B,1B,2D), V(DC,6E,6E,B2), V(B4,5A,5A,EE), V(5B,A0,A0,FB), \
    V(A4,52,52,F6), V(76,3B,3B,4D), V(B7,D6,D6,61), V(7D,B3,B3,CE), \
    V(52,29,29,7B), V(DD,E3,E3,3E), V(5E,2F,2F,71), V(13,84,84,97), \
    V(A6,53,53,F5), V(B9,D1,D1,68), V(00,00,00,00), V(C1,ED,ED,2C), \
    V(40,20,20,60), V(E3,FC,FC,1F), V(79,B1,B1,C8), V(B6,5B,5B,ED), \
    V(D4,6A,6A,BE), V(8D,CB,CB,46), V(67,BE,BE,D9), V(72,39,39,4B), \
    V(94,4A,4A,DE), V(98,4C,4C,D4), V(B0,58,58,E8), V(85,CF,CF,4A), \
    V(BB,D0,D0,6B), V(C5,EF,EF,2A), V(4F,AA,AA,E5), V(ED,FB,FB,16), \
    V(86,43,43,C5), V(9A,4D,4D,D7), V(66,33,33,55), V(11,85,85,94), \
    V(8A,45,45,CF), V(E9,F9,F9,10), V(04,02,02,06), V(FE,7F,7F,81), \
    V(A0,50,50,F0), V(78,3C,3C,44), V(25,9F,9F,BA), V(4B,A8,A8,E3), \
    V(A2,51,51,F3), V(5D,A3,A3,FE), V(80,40,40,C0), V(05,8F,8F,8A), \
    V(3F,92,92,AD), V(21,9D,9D,BC), V(70,38,38,48), V(F1,F5,F5,04), \
    V(63,BC,BC,DF), V(77,B6,B6,C1), V(AF,DA,DA,75), V(42,21,21,63), \
    V(20,10,10,30), V(E5,FF,FF,1A), V(FD,F3,F3,0E), V(BF,D2,D2,6D), \
    V(81,CD,CD,4C), V(18,0C,0C,14), V(26,13,13,35), V(C3,EC,EC,2F), \
    V(BE,5F,5F,E1), V(35,97,97,A2), V(88,44,44,CC), V(2E,17,17,39), \
    V(93,C4,C4,57), V(55,A7,A7,F2), V(FC,7E,7E,82), V(7A,3D,3D,47), \
    V(C8,64,64,AC), V(BA,5D,5D,E7), V(32,19,19,2B), V(E6,73,73,95), \
    V(C0,60,60,A0), V(19,81,81,98), V(9E,4F,4F,D1), V(A3,DC,DC,7F), \
    V(44,22,22,66), V(54,2A,2A,7E), V(3B,90,90,AB), V(0B,88,88,83), \
    V(8C,46,46,CA), V(C7,EE,EE,29), V(6B,B8,B8,D3), V(28,14,14,3C), \
    V(A7,DE,DE,79), V(BC,5E,5E,E2), V(16,0B,0B,1D), V(AD,DB,DB,76), \
    V(DB,E0,E0,3B), V(64,32,32,56), V(74,3A,3A,4E), V(14,0A,0A,1E), \
    V(92,49,49,DB), V(0C,06,06,0A), V(48,24,24,6C), V(B8,5C,5C,E4), \
    V(9F,C2,C2,5D), V(BD,D3,D3,6E), V(43,AC,AC,EF), V(C4,62,62,A6), \
    V(39,91,91,A8), V(31,95,95,A4), V(D3,E4,E4,37), V(F2,79,79,8B), \
    V(D5,E7,E7,32), V(8B,C8,C8,43), V(6E,37,37,59), V(DA,6D,6D,B7), \
    V(01,8D,8D,8C), V(B1,D5,D5,64), V(9C,4E,4E,D2), V(49,A9,A9,E0), \
    V(D8,6C,6C,B4), V(AC,56,56,FA), V(F3,F4,F4,07), V(CF,EA,EA,25), \
    V(CA,65,65,AF), V(F4,7A,7A,8E), V(47,AE,AE,E9), V(10,08,08,18), \
    V(6F,BA,BA,D5), V(F0,78,78,88), V(4A,25,25,6F), V(5C,2E,2E,72), \
    V(38,1C,1C,24), V(57,A6,A6,F1), V(73,B4,B4,C7), V(97,C6,C6,51), \
    V(CB,E8,E8,23), V(A1,DD,DD,7C), V(E8,74,74,9C), V(3E,1F,1F,21), \
    V(96,4B,4B,DD), V(61,BD,BD,DC), V(0D,8B,8B,86), V(0F,8A,8A,85), \
    V(E0,70,70,90), V(7C,3E,3E,42), V(71,B5,B5,C4), V(CC,66,66,AA), \
    V(90,48,48,D8), V(06,03,03,05), V(F7,F6,F6,01), V(1C,0E,0E,12), \
    V(C2,61,61,A3), V(6A,35,35,5F), V(AE,57,57,F9), V(69,B9,B9,D0), \
    V(17,86,86,91), V(99,C1,C1,58), V(3A,1D,1D,27), V(27,9E,9E,B9), \
    V(D9,E1,E1,38), V(EB,F8,F8,13), V(2B,98,98,B3), V(22,11,11,33), \
    V(D2,69,69,BB), V(A9,D9,D9,70), V(07,8E,8E,89), V(33,94,94,A7), \
    V(2D,9B,9B,B6), V(3C,1E,1E,22), V(15,87,87,92), V(C9,E9,E9,20), \
    V(87,CE,CE,49), V(AA,55,55,FF), V(50,28,28,78), V(A5,DF,DF,7A), \
    V(03,8C,8C,8F), V(59,A1,A1,F8), V(09,89,89,80), V(1A,0D,0D,17), \
    V(65,BF,BF,DA), V(D7,E6,E6,31), V(84,42,42,C6), V(D0,68,68,B8), \
    V(82,41,41,C3), V(29,99,99,B0), V(5A,2D,2D,77), V(1E,0F,0F,11), \
    V(7B,B0,B0,CB), V(A8,54,54,FC), V(6D,BB,BB,D6), V(2C,16,16,3A)

/* #define V(a,b,c,d) 0x##a##b##c##d */
#define V(a,b,c,d) 0x##d##c##b##a
#ifndef __HOST__
static uint32_t FT0[256] ALIGN(8) SECTION("section_ft0"); /* 1 KiB @ 0x5000 */
#endif
static uint32_t FT0[256] = { FT };
#undef V

/* #define V(a,b,c,d) 0x##d##a##b##c */
#define V(a,b,c,d) 0x##c##b##a##d
#ifndef __HOST__
static uint32_t FT1[256] ALIGN(8) SECTION("section_ft1"); /* 1 KiB @ 0x5400 */
#endif
static uint32_t FT1[256] = { FT };
#undef V

/* #define V(a,b,c,d) 0x##c##d##a##b */
#define V(a,b,c,d) 0x##b##a##d##c
#ifndef __HOST__
static uint32_t FT2[256] ALIGN(8) SECTION("section_ft2"); /* 1 KiB @ 0x5800 */
#endif
static uint32_t FT2[256] = { FT };
#undef V

/* #define V(a,b,c,d) 0x##b##c##d##a */
#define V(a,b,c,d) 0x##a##d##c##b
#ifndef __HOST__
static uint32_t FT3[256] ALIGN(8) SECTION("section_ft3"); /* 1 KiB @ 0x5c00 */
#endif
static uint32_t FT3[256] = { FT };
#undef V

#undef FT


#define FSbData                                         \
  {                                                     \
    0x63, 0x7C, 0x77, 0x7B, 0xF2, 0x6B, 0x6F, 0xC5,     \
    0x30, 0x01, 0x67, 0x2B, 0xFE, 0xD7, 0xAB, 0x76,     \
    0xCA, 0x82, 0xC9, 0x7D, 0xFA, 0x59, 0x47, 0xF0,     \
    0xAD, 0xD4, 0xA2, 0xAF, 0x9C, 0xA4, 0x72, 0xC0,     \
    0xB7, 0xFD, 0x93, 0x26, 0x36, 0x3F, 0xF7, 0xCC,     \
    0x34, 0xA5, 0xE5, 0xF1, 0x71, 0xD8, 0x31, 0x15,     \
    0x04, 0xC7, 0x23, 0xC3, 0x18, 0x96, 0x05, 0x9A,     \
    0x07, 0x12, 0x80, 0xE2, 0xEB, 0x27, 0xB2, 0x75,     \
    0x09, 0x83, 0x2C, 0x1A, 0x1B, 0x6E, 0x5A, 0xA0,     \
    0x52, 0x3B, 0xD6, 0xB3, 0x29, 0xE3, 0x2F, 0x84,     \
    0x53, 0xD1, 0x00, 0xED, 0x20, 0xFC, 0xB1, 0x5B,     \
    0x6A, 0xCB, 0xBE, 0x39, 0x4A, 0x4C, 0x58, 0xCF,     \
    0xD0, 0xEF, 0xAA, 0xFB, 0x43, 0x4D, 0x33, 0x85,     \
    0x45, 0xF9, 0x02, 0x7F, 0x50, 0x3C, 0x9F, 0xA8,     \
    0x51, 0xA3, 0x40, 0x8F, 0x92, 0x9D, 0x38, 0xF5,     \
    0xBC, 0xB6, 0xDA, 0x21, 0x10, 0xFF, 0xF3, 0xD2,     \
    0xCD, 0x0C, 0x13, 0xEC, 0x5F, 0x97, 0x44, 0x17,     \
    0xC4, 0xA7, 0x7E, 0x3D, 0x64, 0x5D, 0x19, 0x73,     \
    0x60, 0x81, 0x4F, 0xDC, 0x22, 0x2A, 0x90, 0x88,     \
    0x46, 0xEE, 0xB8, 0x14, 0xDE, 0x5E, 0x0B, 0xDB,     \
    0xE0, 0x32, 0x3A, 0x0A, 0x49, 0x06, 0x24, 0x5C,     \
    0xC2, 0xD3, 0xAC, 0x62, 0x91, 0x95, 0xE4, 0x79,     \
    0xE7, 0xC8, 0x37, 0x6D, 0x8D, 0xD5, 0x4E, 0xA9,     \
    0x6C, 0x56, 0xF4, 0xEA, 0x65, 0x7A, 0xAE, 0x08,     \
    0xBA, 0x78, 0x25, 0x2E, 0x1C, 0xA6, 0xB4, 0xC6,     \
    0xE8, 0xDD, 0x74, 0x1F, 0x4B, 0xBD, 0x8B, 0x8A,     \
    0x70, 0x3E, 0xB5, 0x66, 0x48, 0x03, 0xF6, 0x0E,     \
    0x61, 0x35, 0x57, 0xB9, 0x86, 0xC1, 0x1D, 0x9E,     \
    0xE1, 0xF8, 0x98, 0x11, 0x69, 0xD9, 0x8E, 0x94,     \
    0x9B, 0x1E, 0x87, 0xE9, 0xCE, 0x55, 0x28, 0xDF,     \
    0x8C, 0xA1, 0x89, 0x0D, 0xBF, 0xE6, 0x42, 0x68,     \
    0x41, 0x99, 0x2D, 0x0F, 0xB0, 0x54, 0xBB, 0x16      \
  }

#ifndef __HOST__
static uint32_t FSb[256] ALIGN(8) SECTION("section_fsb"); /* 1 KiB @ 0x2000 */
#endif
static uint32_t FSb[256] = FSbData;
#undef FSbData

#define AES_ROUND1(TAB,I,X0,X1,X2,X3,Y0,Y1,Y2,Y3)        \
  {                                                      \
    X0  =      FT0[( Y0       ) &0xFF ];                 \
    X0 ^= rotr(FT0[( Y1 >>  8 ) &0xFF ],24);             \
    X0 ^= rotr(FT0[( Y2 >> 16 ) &0xFF ],16);             \
    X0 ^= rotr(FT0[( Y3 >> 24 ) &0xFF ],8);              \
                                                         \
    X1  =      FT0[( Y1       ) &0xFF ];                 \
    X1 ^= rotr(FT0[( Y2 >>  8 ) &0xFF ],24);             \
    X1 ^= rotr(FT0[( Y3 >> 16 ) &0xFF ],16);             \
    X1 ^= rotr(FT0[( Y0 >> 24 ) &0xFF ],8);              \
                                                         \
    X2  =      FT0[( Y2       ) &0xFF ];                 \
    X2 ^= rotr(FT0[( Y3 >>  8 ) &0xFF ],24);             \
    X2 ^= rotr(FT0[( Y0 >> 16 ) &0xFF ],16);             \
    X2 ^= rotr(FT0[( Y1 >> 24 ) &0xFF ],8);              \
                                                         \
    X3  =      FT0[( Y3       ) &0xFF ];                 \
    X3 ^= rotr(FT0[( Y0 >>  8 ) &0xFF ],24);             \
    X3 ^= rotr(FT0[( Y1 >> 16 ) &0xFF ],16);             \
    X3 ^= rotr(FT0[( Y2 >> 24 ) &0xFF ],8);              \
                                                         \
    X0 ^= (TAB[I++]);                                    \
    X1 ^= (TAB[I++]);                                    \
    X2 ^= (TAB[I++]);                                    \
    X3 ^= (TAB[I++]);                                    \
  }

#define AES_ROUND2(TAB,I,X0,X1,X2,X3,Y0,Y1,Y2,Y3)       \
  {                                                     \
    X0  =      FT0[( Y0       ) &0xFF ];                \
    X0 ^=      FT1[( Y1 >>  8 ) &0xFF ];                \
    X0 ^= rotr(FT0[( Y2 >> 16 ) &0xFF ],16);            \
    X0 ^= rotr(FT1[( Y3 >> 24 ) &0xFF ],16);            \
                                                        \
    X1  =      FT0[( Y1       ) &0xFF ];                \
    X1 ^=      FT1[( Y2 >>  8 ) &0xFF ];                \
    X1 ^= rotr(FT0[( Y3 >> 16 ) &0xFF ],16);            \
    X1 ^= rotr(FT1[( Y0 >> 24 ) &0xFF ],16);            \
                                                        \
    X2  =      FT0[( Y2       ) &0xFF ];                \
    X2 ^=      FT1[( Y3 >>  8 ) &0xFF ];                \
    X2 ^= rotr(FT0[( Y0 >> 16 ) &0xFF ],16);            \
    X2 ^= rotr(FT1[( Y1 >> 24 ) &0xFF ],16);            \
                                                        \
    X3  =      FT0[( Y3       ) &0xFF ];                \
    X3 ^=      FT1[( Y0 >>  8 ) &0xFF ];                \
    X3 ^= rotr(FT0[( Y1 >> 16 ) &0xFF ],16);            \
    X3 ^= rotr(FT1[( Y2 >> 24 ) &0xFF ],16);            \
                                                        \
    X0 ^= (TAB[I++]);                                   \
    X1 ^= (TAB[I++]);                                   \
    X2 ^= (TAB[I++]);                                   \
    X3 ^= (TAB[I++]);                                   \
  }
#define AES_ROUND4(TAB,I,X0,X1,X2,X3,Y0,Y1,Y2,Y3)    \
  {                                                     \
    X0  =      FT0[( Y0       ) &0xFF ];                \
    X0 ^=      FT1[( Y1 >>  8 ) &0xFF ];                \
    X0 ^=      FT2[( Y2 >> 16 ) &0xFF ];                \
    X0 ^=      FT3[( Y3 >> 24 ) &0xFF ];                \
                                                        \
    X1  =      FT0[( Y1       ) &0xFF ];                \
    X1 ^=      FT1[( Y2 >>  8 ) &0xFF ];                \
    X1 ^=      FT2[( Y3 >> 16 ) &0xFF ];                \
    X1 ^=      FT3[( Y0 >> 24 ) &0xFF ];                \
                                                        \
    X2  =      FT0[( Y2       ) &0xFF ];                \
    X2 ^=      FT1[( Y3 >>  8 ) &0xFF ];                \
    X2 ^=      FT2[( Y0 >> 16 ) &0xFF ];                \
    X2 ^=      FT3[( Y1 >> 24 ) &0xFF ];                \
                                                        \
    X3  =      FT0[( Y3       ) &0xFF ];                \
    X3 ^=      FT1[( Y0 >>  8 ) &0xFF ];                \
    X3 ^=      FT2[( Y1 >> 16 ) &0xFF ];                \
    X3 ^=      FT3[( Y2 >> 24 ) &0xFF ];                \
                                                        \
    X0 ^= (TAB[I++]);                                   \
    X1 ^= (TAB[I++]);                                   \
    X2 ^= (TAB[I++]);                                   \
    X3 ^= (TAB[I++]);                                   \
  }

#ifdef __HOST__
/* x86;
   fastest is 4 tables */
#define NFT 4
#else
/* Epiphany:
   fastest is 4 tables, according to both measurement &
   study with ezetime */
#define NFT 4
#endif

#ifndef NFT
#error "must define NFT"
#else
#if (NFT == 1)
#define  AES_ROUND(a,b,Y0,Y1,Y2,Y3,X0,X1,X2,X3) \
         AES_ROUND1(a,b,Y0,Y1,Y2,Y3,X0,X1,X2,X3)
#endif
#if (NFT == 2)
#define  AES_ROUND(a,b,Y0,Y1,Y2,Y3,X0,X1,X2,X3) \
         AES_ROUND2(a,b,Y0,Y1,Y2,Y3,X0,X1,X2,X3)
#endif
#if (NFT == 4)
#define  AES_ROUND(a,b,Y0,Y1,Y2,Y3,X0,X1,X2,X3) \
         AES_ROUND4(a,b,Y0,Y1,Y2,Y3,X0,X1,X2,X3)
#endif
#endif

static inline void aes256_encrypt(const uint32_t *input, uint32_t *output, const uint32_t *aes_edrk )
{
  uint32_t X0, X1, X2, X3, Y0, Y1, Y2, Y3;
  unsigned int i = 0, j = 0;
  unsigned int l_aes_nr = 14;

  X0 = (input[i++] ^ aes_edrk[j++]);
  X1 = (input[i++] ^ aes_edrk[j++]);
  X2 = (input[i++] ^ aes_edrk[j++]);
  X3 = (input[i++] ^ aes_edrk[j++]);

/*   printf("%d: 0x%08x 0x%08x 0x%08x 0x%08x\n", 4, X0, X1, X2, X3); */

  for (i = 4 ; i < (l_aes_nr<<2) ; )
  {
    
    AES_ROUND(aes_edrk, i, Y0, Y1, Y2, Y3, X0, X1, X2, X3 );

    X0=Y0;
    X1=Y1;
    X2=Y2;
    X3=Y3;
  }
  /* last round */
  Y0 = (aes_edrk[i]) ^
    ( FSb[( X0       ) &0xFF ]       ) ^
    ( FSb[( X1 >>  8 ) &0xFF ] <<  8 ) ^
    ( FSb[( X2 >> 16 ) &0xFF ] << 16 ) ^
    ( FSb[( X3 >> 24 ) &0xFF ] << 24 );
  
  Y1 = (aes_edrk[1+i]) ^
    ( FSb[( X1       ) &0xFF ]       ) ^
    ( FSb[( X2 >>  8 ) &0xFF ] <<  8 ) ^
    ( FSb[( X3 >> 16 ) &0xFF ] << 16 ) ^
    ( FSb[( X0 >> 24 ) &0xFF ] << 24 );
  
  Y2 = (aes_edrk[2+i]) ^
    ( FSb[( X2       ) &0xFF ]       ) ^
    ( FSb[( X3 >>  8 ) &0xFF ] <<  8 ) ^
    ( FSb[( X0 >> 16 ) &0xFF ] << 16 ) ^
    ( FSb[( X1 >> 24 ) &0xFF ] << 24 );
  
  Y3 = (aes_edrk[3+i]) ^
    ( FSb[( X3       ) &0xFF ]       ) ^
    ( FSb[( X0 >>  8 ) &0xFF ] <<  8 ) ^
    ( FSb[( X1 >> 16 ) &0xFF ] << 16 ) ^
    ( FSb[( X2 >> 24 ) &0xFF ] << 24 );

  output[0] = (Y0);
  output[1] = (Y1);
  output[2] = (Y2);
  output[3] = (Y3);
}

#ifdef __HOST__
static inline int aes256_set_key(const uint32_t key[], uint32_t *aes_edrk)
{
    unsigned int i = 0;
    uint32_t rotl_aes_edrk;
      uint32_t aes_edrk_temp8, aes_edrk_temp9, aes_edrk_temp10, aes_edrk_temp11;
      uint32_t aes_edrk_temp12, aes_edrk_temp13, aes_edrk_temp14, aes_edrk_temp15;
      uint32_t temp_lds;
      uint32_t round = 0x00000001;

    aes_edrk_temp8  = (key[0]);
    aes_edrk[0] = aes_edrk_temp8;
    aes_edrk_temp9  = (key[1]);
    aes_edrk[1] = aes_edrk_temp9;
    aes_edrk_temp10 = (key[2]);
    aes_edrk[2] = aes_edrk_temp10;
    aes_edrk_temp11 = (key[3]);
    aes_edrk[3] = aes_edrk_temp11;
    aes_edrk_temp12 = (key[4]);
    aes_edrk[4] = aes_edrk_temp12;
    aes_edrk_temp13 = (key[5]);
    aes_edrk[5] = aes_edrk_temp13;
    aes_edrk_temp14 = (key[6]);
    aes_edrk[6] = aes_edrk_temp14;
    aes_edrk_temp15 = (key[7]);
    aes_edrk[7] = aes_edrk_temp15;

    for( i = 8; i < 64; /* i+=8 */ )
    {
      rotl_aes_edrk   = rotr(aes_edrk_temp15,8);

      temp_lds = f_FSb_32__1(rotl_aes_edrk) ^ f_FSb_32__2( rotl_aes_edrk );

      aes_edrk_temp8 = aes_edrk_temp8 ^ round ^ temp_lds;
      round = round << 1;

      aes_edrk[i++]   = aes_edrk_temp8;
      aes_edrk_temp9  = aes_edrk_temp9  ^ aes_edrk_temp8;
      aes_edrk[i++]   = aes_edrk_temp9;
      aes_edrk_temp10 = aes_edrk_temp10 ^ aes_edrk_temp9;
      aes_edrk[i++]  = aes_edrk_temp10;
      aes_edrk_temp11 = aes_edrk_temp11 ^ aes_edrk_temp10;
      aes_edrk[i++]  = aes_edrk_temp11;

      temp_lds = f_FSb_32__1(aes_edrk_temp11) ^ f_FSb_32__2(aes_edrk_temp11);

      aes_edrk_temp12 = aes_edrk_temp12 ^ temp_lds;
      aes_edrk[i++]  = aes_edrk_temp12;
      aes_edrk_temp13 = aes_edrk_temp13 ^ aes_edrk_temp12;
      aes_edrk[i++]  = aes_edrk_temp13;
      aes_edrk_temp14 = aes_edrk_temp14 ^ aes_edrk_temp13;
      aes_edrk[i++]  = aes_edrk_temp14;
      aes_edrk_temp15 = aes_edrk_temp15 ^ aes_edrk_temp14;
      aes_edrk[i++]  = aes_edrk_temp15;
    }

    return( 0 );
}

void aes256_prepare_epi_keys(const uint32_t key[], const uint32_t IV[],my_epiphany* epip) {
  uint32_t aes_edrk[64];
  size_t r;
  aes256_set_key(key,aes_edrk);

  E_CHECK_ERR(e_write(&epip->DRAM, 0, 0, offsetof(shared_buf_t, k), (void *)aes_edrk, sizeof(aes_edrk)));  
}


void aes256_host_encrypt(const uint32_t key[], const uint32_t IV[], const uint8_t* in, uint8_t *out, size_t size) {
  uint32_t aes_edrk[64];
  aes256_set_key(key,aes_edrk);
  for (size_t i = 0 ; i < (size+15)/16 ; i++) {
    uint32_t buf[4];
    uint32_t t[4] = {0, 0, 0, 0};
    t[3] = __builtin_bswap32(i&0xFFFFFFFF);
/*     if (sizeof(size_t)>4) */
/*       t[2] = __builtin_bswap32((i>>32)&0xFFFFFFFF); */
    aes256_encrypt(t, buf, aes_edrk);
    size_t start = i*16, stop = i*16+16;
    if (stop > size)
      stop = size;
    for (size_t j = start ; j < stop ; j++) {
      out[j] = in[j] ^ ((uint8_t*)buf)[j-start];
    }
  }
}


void aes256_openssl_encrypt(const uint32_t key[], const uint32_t IV[], const uint8_t* in, uint8_t *out, size_t size) {
  EVP_CIPHER_CTX x;
  int ok = 1;
  int outl = 0;
  EVP_CIPHER_CTX_init(&x);
  ok = EVP_EncryptInit_ex(&x,EVP_aes_256_ctr(),0,(const uint8_t*)key,(const uint8_t*)IV);
  if (ok == 1)
    ok = EVP_CIPHER_CTX_set_padding(&x, 0);
  if (ok == 1)
    ok = EVP_EncryptUpdate(&x, (uint8_t*)out, &outl, (const uint8_t*)in, size);
  if (ok == 1)
    ok = EVP_EncryptFinal_ex(&x, (uint8_t*)out, &outl);
  if (!ok)
    memset(out, 0, size);
}
#endif

#ifndef __HOST__

int main(int argc, char **argv) {

  init();

  while (1) {
    while (Mailbox.pCore->cn[me.corenum] == 0) {
      /* wait */
    }
    uint64_t cs = Mailbox.pCore->cs[me.corenum];
    uint64_t cn = Mailbox.pCore->cn[me.corenum];
    //uint64_t cd = Mailbox.pCore->cd[me.corenum];
    uint64_t mcd = 0;

#define ISOLATE_(X) \
    for (int i = 0 ; i < 16 ; i++) {            \
      e_barrier(barriers, tgt_bars);            \
      if (me.corenum == i)                      \
        X;                                      \
    }
#define ISOLATE(X) X;
    
    ISOLATE(e_dma_copy(lk, Mailbox.k, 256))
    
    ISOLATE(Mailbox.pCore->cn[me.corenum] = 0)

    e_irq_attach(E_USER_INT, empty_isr);
    
    for (uint64_t l = 0 ; l < cn ; l += PCSSIZE/16) {
      uint64_t scount = 0;
      for (uint64_t j = 0 ; j < PCSSIZE/16 && (j+l) < cn ; j += BSIZE/16) {
        uint64_t count = 0;
        for (uint64_t k = 0 ; k < BSIZE/16 && (k+j+l) < cn; k++) {
          uint64_t i = l + j + k;
          uint64_t ctr = cs + i;
          uint32_t ctra[4];
          ctra[0] = 0;
          ctra[1] = 0;
          ctra[2] = __builtin_bswap32((uint32_t)(0xFFFFFFFF & (ctr>>32)));
          ctra[3] = __builtin_bswap32((uint32_t)(0xFFFFFFFF & (ctr    )));
          aes256_encrypt(ctra, lc+(4*k), lk);
          count++;
        }
        ISOLATE(e_dma_copy(Mailbox.c+(me.corenum*(PCSSIZE/sizeof(uint32_t)))+(scount*4), lc, 16*count))
        mcd += count;
        scount += count;
        ISOLATE(Mailbox.pCore->cd[me.corenum] = mcd)
      }
      wait_on_user();
    }
  }

  return 0;
}
#endif // __HOST__
