/*
Copyright (c) 2014, Romain Dolbeau, unless otherwise noted.
No claims is made upon work by others.

For the work by Romain Dolbeau:
All rights reserved

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

For the work by others:
See the indicated reference for the relevant license.
*/

#include "epi_crypto_common.h"

#include "e_epi_crypto.h"

#include <e_lib.h>

vuint32_t lk[64]   ALIGN(8) SECTION("section_key"); /* 256 B @ 0x4000 */
/* vuint32_t lm[1024] ALIGN(8) SECTION("section_m"); */ /* up to 4 KiB @ 0x2000 */
vuint32_t lc[BSIZE/sizeof(uint32_t)] ALIGN(8) SECTION("section_c"); /* up to 8 KiB @ 0x2000 */
core_t           me        ALIGN(8) SECTION("core_data_section"); /* @ 0x7000+ */
shared_buf_ptr_t Mailbox   ALIGN(8) SECTION("core_data_section"); /* @ 0x7000+ */


volatile e_barrier_t  barriers[16] SECTION("core_data_section"); // barriers array
         e_barrier_t *tgt_bars[16] SECTION("core_data_section"); // barriers array


extern const unsigned _SHARED_DRAM_;
#define SHARED_DRAM   ((unsigned)(&_SHARED_DRAM_))

void setIntegerMode(uint32_t e) {
  uint32_t r, m;
  __asm__ volatile("MOVFS %0,CONFIG" : "=r"(r));
  m = 0xFFFFFFFF ^ ((1-e) << 19);
  r = (r | (1u<<19)) & m;
  __asm__ volatile("MOVTS CONFIG,%0" : : "r"(r));
}

void wait_on_user(void) {
#ifdef SYNCHRO_BY_SIGNAL
  uint32_t m, mo;
  __asm__ volatile("MOVFS %0,IMASK" : "=r"(mo));
  m = (mo | 0x000001FF) & ~(1u<<9);
  __asm__ volatile("MOVTS IMASK,%0" : : "r"(m));
  __asm__ volatile("IDLE");
  __asm__ volatile("MOVTS IMASK,%0" : : "r"(mo));
#else
  while (Mailbox.pCore->cd[me.corenum] != 0) {
    /* wait */
  }
#endif
}

void __attribute__((interrupt)) empty_isr()
{
        return;
}

void init(void) {
  Mailbox.pBase = (void *) SHARED_DRAM;
  Mailbox.c    = Mailbox.pBase + offsetof(shared_buf_t, c);
  Mailbox.k    = Mailbox.pBase + offsetof(shared_buf_t, k);
  Mailbox.pCore = Mailbox.pBase + offsetof(shared_buf_t, core);
  me.row = e_group_config.core_row;
  me.col = e_group_config.core_col;
  me.corenum = me.row * e_group_config.group_rows + me.col;
  Mailbox.pCore->cn[me.corenum] = 0;
  Mailbox.pCore->cs[me.corenum] = 0;
  Mailbox.pCore->cd[me.corenum] = 0;
  e_barrier_init(barriers, tgt_bars);
  e_barrier(barriers, tgt_bars);
  if (me.corenum == 0) {
    Mailbox.pCore->ready = 1;
#if 0
    // try accessing the FPGA GCM registers ; testing stuff for epiphany -> fpga communications
    vuint32_t *gcm_base_addr = (vuint32_t*)(((vuint8_t*)SHARED_DRAM)+0x01000000);
    int i;
    for (i = -1 ; i < 2 ; i++)
      (*(gcm_base_addr+i)) = 0xBADC0FFE;
#endif
  }
}
