#ifndef _EPI_CRYPTO_COMMON_H_
#define _EPI_CRYPTO_COMMON_H_

#include <stdint.h>

/* can't try with anything other than 16 */
#define NCORE 16
/* how much room we have ; AES tables take a lot of space...
   current limit is 8192 (easy to push to 12288 */
#define BSIZE 8192
/* unreliable ? */
//#define SYNCHRO_BY_SIGNAL
#define PPCINS 16
#define PCSSIZE (BSIZE*PPCINS)
#define SSIZE (NCORE*PCSSIZE)

typedef volatile uint8_t vuint8_t;
typedef volatile uint16_t vuint16_t;
typedef volatile uint32_t vuint32_t;
typedef volatile uint64_t vuint64_t;

typedef struct {
  uint32_t        corenum;
  uint32_t        row;
  uint32_t        col;
  uint32_t        mystate;
} core_t;

typedef struct {
  vuint64_t cs[16];
  vuint64_t cn[16];
  vuint64_t cd[16];
  vuint32_t start;  /* AES block */
  vuint32_t number; /* AES block */
  vuint32_t ready;
  vuint32_t pad;
} mbox_t;

typedef struct {
  volatile mbox_t core;
  vuint32_t k[64];    /* fit AES round keys ... */
  vuint32_t IV[4];
  vuint32_t c[SSIZE/sizeof(uint32_t)];
} shared_buf_t;

/* #define rotr(x,n)   (((x) >> ((int)(n))) | ((x) << (32 - (int)(n)))) */
/* #define rotl(x,n)   (((x) << ((int)(n))) | ((x) >> (32 - (int)(n)))) */
static inline uint32_t rotr(const uint32_t x, const uint32_t n) {
  uint32_t r;
  r = ((x >> n) | (x << (32 - n)));
  return r;
}
static inline uint32_t rotl(const uint32_t x, const uint32_t n) {
  uint32_t r;
  r = ((x << n) | (x >> (32 - n)));
  return r;
}

#ifdef __HOST__
#include <stdio.h>
#include <stddef.h>
#include <e-hal.h>

typedef struct {
  e_platform_t platform;
  e_mem_t DRAM;
  e_epiphany_t dev;

  volatile void *ptr;
  int fd;
  unsigned int page_offset;
} my_epiphany;

#define E_CHECK_ERR(C) r = C; if (r == E_ERR) fprintf(stderr, "%s:%d: %s returned %d\n", __PRETTY_FUNCTION__, __LINE__, #C, r)

typedef void (*prepare_epi_keys)(const uint32_t key[], const uint32_t IV[], my_epiphany* epip);

typedef void (*host_encrypt)(const uint32_t key[], const uint32_t IV[], const uint8_t* in, uint8_t *out, size_t size);
#endif

#endif // _EPI_CRYPTO_COMMON_H_
